{-
  This module defines the basic laws of physics.
-}

module Laws (
  move,
  nextState
) where

import Planet
import Control.Parallel.Strategies
import Control.Parallel(par, pseq)

type Force = Float
type Acc = Float
type Vel = Float
type Time = Float

{-
  Gravitational constant.
-}
g :: Float
g = 10

{-
  This function takes two Planets as input, computes
  the gravity from p1 to p2, and returns a
  list of float value which represents Fx and Fy.
  Gravity formula F = g * m1 * m2 / r^2
-}
gravity :: Planet -> Planet -> [Force]
gravity p1 p2 = [if x2 > x1 then fx else -fx, if y2 > y1 then fy else -fy]
  where fx = if x1 == x2 then 0 
              else g * m1 * m2 / rSqr
        fy = if y1 == y2 then 0
              else g * m1 * m2 / rSqr
        x1 = posX p1
        x2 = posX p2
        y1 = posY p1
        y2 = posY p2
        m1 = mass p1
        m2 = mass p2
        rSqr = (fSqr (x1 - x2)) + (fSqr (y1 - y2))

{-
  Takes a planet and a pair of forces.
  Returns a pair of velocities.
-}
acceleration :: Planet -> [Force] -> [Acc]
acceleration p fxy = map acc fxy
  where acc f = f / (mass p)

{-
  Change velocities.
-}
accelerate :: [Vel] -> [Acc] -> Time -> [Vel]
accelerate vs as t = zipWith (\v a -> v + a * t) vs as

{-
  Used for visualization.
-}
nextState :: [Planet] -> Double -> [Planet]
nextState ps _ = [move p ps (1/100) | p <- ps] `using` parList rseq

move :: Planet -> [Planet] -> Time -> Planet
-- move p ps t = moveHelper 100 p ps (t / 100)
move p ps t = moveHelper 1 p ps t

moveHelper :: Int -> Planet -> [Planet] -> Time -> Planet
moveHelper 0 p ps t = p
moveHelper n p ps t = (moveHelper (n - 1) p' ps t)
  where p'          = Planet vs m x y
        vs          = accelerate (velocity p) as t
        as          = acceleration p fs
        -- fs          = foldr (zipWith (+)) [0, 0] ((map (gravity p) ps) `using` parList rseq)
        fs          = foldr (zipWith (+)) [0, 0] (map (gravity p) ps)
        [x, y]      = zipWith (\pos v -> pos + v * t) [posX p, posY p] vs
        m           = mass p

fSqr :: Float -> Float
fSqr x = x * x
